"""
Interface en ligne de commande pour flumel:
* config : création assistée du fichier de conf
* page : génération de la page dédiée de l’instance
* init : execute les 3 tâches précédentes à la suite

Pour un premier lancement, utiliser "init"
"""

import argparse


def user_input(label: str = 'Poursuivre (ctr+c pour annuler)',
               default: str = '') -> str:
    """
    Simplifie la récupération des variables pour la configuration

    Affiche : Question [valeur par défaut] ?
    Retourne : Valeur saisie par l’utilisateur OU valeur par défaut

    Utilisé sans paramètre permet de faire une pause dans l'affichage
    et demander à l'utilisateur de poursuivre
    """
    question = '{label} [{default}] ? '.format(label=label, default=default)
    return str(input(question) or default)


def main():
    """init | config | page"""
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('command', help=main.__doc__)
    args = parser.parse_args()
    available_commands = {
        'init': init,
        'config': generate_config,
        'page': generate_page
    }
    try:
        available_commands[args.command]()
    except KeyError:
        print('⚠ Commande inconnue, cli.py --help pour afficher l’aide')


def init():
    """
    Initialisation de l'instance
    Execute generate_config & generate_page
    """
    print(init.__doc__)
    user_input()
    generate_config()
    generate_page()


def generate_config():
    """
    Génération du fichier de configuration de flumel

    * Format des questions: "Question [valeur par défaut] ?"
    * Quand il n’y a pas de valeur par défaut (crochets vides)
    il faut impérativement répondre quelque chose.
    * Une fois généré, le fichier flumel.cfg peut être modifié à la main.
    """
    import configparser
    config = configparser.ConfigParser(interpolation=None)
    config.read('flumel.cfg')
    print(generate_config.__doc__)
    user_input()

    print('# Réglages de l’instance')
    try:
        config.add_section('Instance')
    except configparser.DuplicateSectionError:
        pass
    config.set('Instance', 'Name', user_input('- Nom de l’instance', 'Flumel'))
    config.set('Instance', 'Url', user_input('- URL de l’instance'))
    config.set(
        'Instance', 'Baseline',
        user_input('- Slogan de l’instance',
                   'Encore un moyen de recevoir ses flux RSS par email'))

    print('# Réglages du bot')
    try:
        config.add_section('Bot')
    except configparser.DuplicateSectionError:
        pass
    config.set('Bot', 'Email', user_input('- Adresse email du bot'))

    print('# Réglages IMAP')
    try:
        config.add_section('IMAP')
    except configparser.DuplicateSectionError:
        pass
    config.set('IMAP', 'Host', user_input('- Adresse du serveur IMAP'))
    config.set('IMAP', 'Port',
               user_input('- Port (993 pour SSL, 143 pour StartTLS)', '465'))
    config.set('IMAP', 'Login', user_input('- Identifiant IMAP'))
    config.set('IMAP', 'Password', user_input('- Mot de passe IMAP'))

    print('# Réglages SMTP')
    try:
        config.add_section('SMTP')
    except configparser.DuplicateSectionError:
        pass
    config.set('SMTP', 'Host', user_input('- Adresse du serveur SMTP'))
    config.set('SMTP', 'Port',
               user_input('- Port (465 pour SSL, 587 pour StartTLS) ', '993'))
    config.set('SMTP', 'Login', user_input('- Identifiant SMTP'))
    config.set('SMTP', 'Password', user_input('- Mot de passe SMTP'))

    print('# Commandes')
    try:
        config.add_section('Keywords')
    except configparser.DuplicateSectionError:
        pass
    config.set('Keywords', 'Subscribe',
               user_input('- Pour s’abonner', 'Abonnement'))
    config.set('Keywords', 'Unsubscribe',
               user_input('- Pour se désabonner', 'Désabonnement'))

    print('# Base de données')
    try:
        config.add_section('DB')
    except configparser.DuplicateSectionError:
        pass
    config.set('DB', 'Path',
               user_input('- Nom et chemin de la base', 'flumel.sqlite'))
    config.set('DB', 'Debug', user_input('- Debug activé (yes, no)', 'no'))

    print('# Tâches')
    try:
        config.add_section('Queue')
    except configparser.DuplicateSectionError:
        pass
    config.set('Queue', 'Path',
               user_input('- Nom et chemin de la base', 'huey.sqlite'))
    config.set(
        'Queue', 'Management',
        user_input(
            '- Fréquence de vérification IMAP (*/n = toutes les n minutes)',
            '*/3'))
    config.set(
        'Queue', 'Subscriptions',
        user_input(
            '- Fréquence de vérification des flux (*/n = toutes les n heures)',
            '*/1'))

    print('# Journaux')
    try:
        config.add_section('Logging')
    except configparser.DuplicateSectionError:
        pass
    config.set('Logging', 'Path',
               user_input('- Nom et chemin du journal', 'flumel.log'))
    config.set('Logging', 'Level',
               user_input('- Niveau (DEBUG, INFO, WARN, ERROR, CRITICAL)',
                          'INFO'))

    with open('flumel.cfg', 'w') as fp_config:
        config.write(fp_config)
    print('✓ flumel.cfg généré !')


def generate_page():
    """
    Créer de la page dédiée de l’instance
    web/index.html
    """
    print(generate_page.__doc__)
    user_input()
    from .tasks import page
    page()
    print('✓ index.html créé !')


if __name__ == "__main__":
    main()



"""
Boite à outils
"""
import imaplib
import smtplib
from datetime import datetime
from email import message_from_bytes
from email.header import decode_header, make_header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from string import Template
from textwrap import shorten
from time import mktime
from xml.sax.saxutils import escape, quoteattr

import html2text
import pkg_resources
import requests
from fake_useragent import UserAgent
from premailer import transform
from readability import Document

from . import settings


def format_template(name: str, **kwargs) -> str:
    """
    Formate le contenu du gabarit `name` avec kwargs + un extrait des settings.
    Utilise `string.Template`, les chaînes `$foo` seront remplacées par la valeur de `foo`.
    On essaye d'utiliser le gabarit personnalisé dans le répertoire `templates/`
    Si introuvable, replis sur celui du package.
    Note: pkg_resources.resource_string est utilisé pour récupérer le path de flumel
    """
    kwargs.update({
        'INSTANCE_NAME': settings.INSTANCE_NAME,
        'INSTANCE_URL': settings.INSTANCE_URL,
        'INSTANCE_BASELINE': settings.INSTANCE_BASELINE,
        'BOT_EMAIL': settings.BOT_EMAIL,
        'SUBSCRIBE_KEYWORD': settings.SUBSCRIBE_KEYWORD,
        'UNSUBSCRIBE_KEYWORD': settings.UNSUBSCRIBE_KEYWORD
    })
    path = '/'.join(['templates', name])
    try:
        with open(path, 'r') as fp_template:
            html = fp_template.read()
    except FileNotFoundError:
        html = pkg_resources.resource_string(__name__, path).decode('utf-8')
    return Template(html).safe_substitute(**kwargs)


def decode_email_header(value) -> str:
    """
    Les headers des emails sont encodés comme des sagouins
    https://docs.python.org/3/library/email.header.html#email.header.decode_header
    @see https://stackoverflow.com/a/12071098
    """
    return str(make_header(decode_header(value))).strip()


def escape_xml(text: str) -> str:
    """
    Échappe les caractères spéciaux d’une chaîne pour intégration dans du xml
    Utilisé pour l’export OPML
    """
    return escape(quoteattr(text))


class Article:
    """
    La synthèse des informations contenues dans le flux RSS et dans ce que
    l’on peut récupérer sur la page de l’article :

    * title - str - champ RSS
    * summary - str - champ RSS
    * body - str - extraction site | champ RSS
    * authors - str - champ RSS
    * date - date - champ RSS
    * link - str - champ RSS
    * keywords - str - champ RSS
    """

    title = ''
    _body = ''
    link = ''
    entry = None

    def __init__(self, entry: object):
        """
        Récupération de la page et dispatch des informations
        """
        self.entry = entry
        self.link = entry.link
        self.title = entry.title
        self.summary = shorten(entry.summary, width=64, placeholder='...')

    @property
    def body(self) -> str:
        """
        Tentative de récupération du texte complet de l’article via readability
        Si erreur de récupération de la page on retourne le contenu du flux.
        Si longeur de l’extraction readability > au summary du flux on s'en sert,
        sinon on utilise le flux
        """
        if self._body:  # pseudo cache
            return self._body
        user_agent = UserAgent()
        headers = {'User-Agent': user_agent.ff}
        response = requests.get(self.link, headers=headers)
        if response.status_code != 200:  # lien hs, on retourne le contenu du rss
            self._body = ''
            return self.entry.content[0].value
        try:
            doc = Document(response.text)
            if len(doc.summary()) > len(
                    self.entry.summary):  # extraction > flux
                self._body = doc.summary()
            else:
                self._body = self.entry.summary
            return self._body
        except ValueError:  # erreur dans readability
            self._body = ''
            return self.entry.summary

    @property
    def date(self) -> datetime:
        """
        Retourne la date de publication ou la date de maj
        """
        try:
            return datetime.fromtimestamp(mktime(self.entry.published_parsed))
        except (OverflowError, ValueError):
            return datetime.fromtimestamp(mktime(self.entry.updated_parsed))

    @property
    def authors(self) -> str:
        """
        Ajoute author à contributors et retourne une liste séparée par des virgules
        """
        authors = []
        try:
            authors.append(self.entry.author)
        except AttributeError:
            pass
        try:
            for contributor in self.entry.contributors:
                authors.append(contributor.name)
        except AttributeError:
            pass
        return ', '.join(authors)

    @property
    def keywords(self) -> str:
        """
        Retourne une liste séparée par des virgules des étiquettes / categories
        """
        keywords = []
        try:
            for tag in self.entry.tags:
                keywords.append(tag.label if tag.label else tag.term)
        except AttributeError:
            pass
        return ', '.join(keywords)


class IMAP:
    """
    Interface simplifiée pour imaplib.IMAP4
    La connexion est en SSL si BOT_IMAP_PORT == 993, sinon c'est en starttls
    """

    server = None

    def __init__(self):
        """
        Connexion & login
        """
        if settings.BOT_IMAP_PORT == 993:
            self.server = imaplib.IMAP4_SSL(settings.BOT_IMAP_HOST)
        else:
            self.server = imaplib.IMAP4(settings.BOT_IMAP_HOST)
            self.server.starttls()
        self.server.login(settings.BOT_IMAP_LOGIN, settings.BOT_IMAP_PASSWORD)

    @property
    def inbox(self) -> list:
        """
        Récupération de la liste des messages dans l’inbox, retourne un tableau iterable
        """
        self.server.select()  # ouverture inbox
        _, listing = self.server.search(None, 'ALL')  # (résultat, list)
        return listing[0].split()

    def get(self, number: int):
        """
        Retourne le message `number`
        """
        _, msg = self.server.fetch(number, '(RFC822)')  # (résultat, message)
        return message_from_bytes(msg[0][1])

    def delete(self, number: int):
        """
        Flag le message `number` à supprimer
        """
        return self.server.store(number, '+FLAGS', '\\Deleted')

    def purge(self):
        """
        Vide l’inbox après avoir marqué l’ensemble des messages à supprimer
        """
        for number in self.inbox:
            self.delete(number)
        return self.server.expunge()

    def close(self) -> None:
        """
        Efface les messages avec flag et clôture la connexion
        """
        self.server.select()  # command EXPUNGE only allowed in states SELECTED
        self.server.expunge()
        self.server.close()
        self.server.logout()


class SMTP:
    """
    Interface simplifiée pour smtplib.SMTP
    La connexion est en SSL si BOT_SMTP_PORT == 465, sinon c'est en starttls
    """

    server = None

    def __init__(self):
        """
        Connexion & login
        """
        if settings.BOT_SMTP_PORT == 465:
            self.server = smtplib.SMTP_SSL(settings.BOT_SMTP_HOST,
                                           settings.BOT_SMTP_PORT)
        else:
            self.server = smtplib.SMTP(settings.BOT_SMTP_HOST,
                                       settings.BOT_SMTP_PORT)
            if settings.BOT_SMTP_PORT == 587:
                self.server.starttls()
        try:
            self.server.login(settings.BOT_SMTP_LOGIN,
                              settings.BOT_SMTP_PASSWORD)
        except smtplib.SMTPNotSupportedError:
            pass

    def send(self,
             recipient: str,
             subject: str,
             html: str,
             preheader: str = None,
             **kwargs) -> None:
        """
        Envoi d'un email multipart intégré dans des gabarits
        Le corps du message est envoyé en html et sa conversion en texte brut par html2text
        Les kwargs permettent de personnaliser les entêtes du mail X-Flumel
        par ex. Title="Test" -> X-Flumel-Title = "Test"
        """

        message = MIMEMultipart('alternative')
        message['Subject'] = subject
        message['From'] = settings.BOT_EMAIL
        message['To'] = recipient
        if kwargs:
            for key, value in kwargs.items():
                message['X-Flumel-' + key] = value

        text = html2text.html2text(html)
        full_txt = format_template('email.txt', **{
            'content': text,
        })
        message.attach(MIMEText(full_txt, 'plain'))

        full_html = format_template('email.html', **{
            'content': html,
            'preheader': preheader,
            'title': subject
        })
        inline_html = transform(full_html)
        message.attach(MIMEText(inline_html, 'html'))

        self.server.sendmail(settings.BOT_EMAIL, [recipient],
                             message.as_string())

    def close(self):
        """
        Clôture de la connexion
        """
        self.server.quit()
